unit VideoPlayer;

interface

uses
  Windows, ActiveX, MMSystem, SysUtils, DShow,
  RiffWalk;

const
  IID_IMediaDet: TGUID = '{65BD0710-24D2-4FF7-9324-ED2E5D3ABAFA}';
  CLSID_MediaDet: TGUID = '{65BD0711-24D2-4FF7-9324-ED2E5D3ABAFA}';

type
  // DShow.pasɂȂC^[tF[X
  ISampleGrabberCB = interface(IUnknown)
    ['{0579154A-2B53-4994-B0D0-E773148EFF85}']
    function SampleCB(SampleTime: Double; pSample: IMediaSample): HResult; stdcall;
    function BufferCB(SampleTime: Double; pBuffer: PByte; BufferLen: longint): HResult; stdcall;
  end;
  ISampleGrabber = interface(IUnknown)
    ['{6B652FFF-11FE-4FCE-92AD-0266B5D7C78F}']
    function SetOneShot(OneShot: BOOL): HResult; stdcall;
    function SetMediaType(var pType: TAM_MEDIA_TYPE): HResult; stdcall;
    function GetConnectedMediaType(out pType: TAM_MEDIA_TYPE): HResult; stdcall;
    function SetBufferSamples(BufferThem: BOOL): HResult; stdcall;
    function GetCurrentBuffer(var pBufferSize: longint; pBuffer: Pointer): HResult; stdcall;
    function GetCurrentSample(out ppSample: IMediaSample): HResult; stdcall;
    function SetCallback(pCallback: ISampleGrabberCB; WhichMethodToCallback: longint): HResult; stdcall;
  end;
  IMediaDet = interface(IUnknown)
    ['{65BD0710-24D2-4FF7-9324-ED2E5D3ABAFA}']
    function get_Filter(out pVal: IUnknown): HResult; stdcall;
    function put_Filter(newVal: IUnknown): HResult; stdcall;
    function get_OutputStreams(out pVal: longint): HResult; stdcall;
    function get_CurrentStream(out pVal: longint): HResult; stdcall;
    function put_CurrentStream(newVal: longint): HResult; stdcall;
    function get_StreamType(out pVal: TGUID): HResult; stdcall;
    function get_StreamTypeB(out pVal: WideString): HResult; stdcall;
    function get_StreamLength(out pVal: Double): HResult; stdcall;
    function get_Filename(out pVal: WideString): HResult; stdcall;
    function put_Filename(pVal: WideString): HResult; stdcall;
    function GetBitmapBits(streamTime: Double; var pBufferSize: longint; var pBuffer: Pchar;
      Width: longint; Height: longint): HResult; stdcall;
    function WriteBitmapBits(streamTime: Double; Width: longint; Height: longint;
      Filename: WideString): HResult; stdcall;
    function get_StreamMediaType(out pVal: TAM_MEDIA_TYPE): HResult; stdcall;
    function GetSampleGrabber(out ppVal: ISampleGrabber): HResult; stdcall;
    function get_FrameRate(out pVal: Double): HResult; stdcall;
    function EnterBitmapGrabMode(SeekTime: Double): HResult; stdcall;
  end;

  TVPModes = (vpPlay, vpPause, vpStop);

  TVideoPlayer = class(TObject)
  protected
    Figb: IGraphBuilder;
    Fimc: IMediaControl;
    Fims: IMediaSeeking;
    Fime: IMediaEvent;
    Fibv: IBasicVideo;
    //Fivf: IVideoFrameStep;
    Fiba: IBasicAudio;
    Fivw: IVideoWindow;
    TimerID: Integer;
    Duration: Comp;
    function GetVolume: Integer;
    procedure SetVolume(Value: Integer);
    function GetPosition: Integer;
    procedure SetPosition(Value: Integer);
    function GetAVIHeader(var F: File): Boolean;
    procedure GetMpegInfo(var F: File; var Aspect: Double; var w, h: Integer);
  public
    Mode: TVpModes;
    FileName: string;
    Width: Integer;
    Height: Integer;
    Length: Integer;   // msec
    Loaded: Boolean;
    DisplayWnd: HWND;
    Fps: Double;
    Aspect: Double;
    AVIMainHeader: TAVIMainHeader;
    BitmapInfoHeader: TBitmapInfoHeader;
    WaveFormat: TWaveFormat;
    FilterNames: array[0..255] of string;
    FilterCount: Integer;
    OnTimer: procedure of object;
    OnEndPlay: procedure of object;
    property Volume: Integer read GetVolume write SetVolume; // -10000..0
    property Position: Integer read GetPosition write SetPosition; // msec
    procedure FreeInterface;
    procedure LoadFile(const FileName: string);
    procedure Play;
    procedure Stop;
    procedure Pause;
    procedure PauseOnly;
    procedure SetWindowPosition(ALeft, ATop, AWidth, AHeight: Longint);
    function WriteBitmapBits(const FileName: string): Boolean;
    procedure GetFilterNames;
    procedure ShowFilterPropertyByName(const FilterName: string);
  end;

implementation

const
  Mpeg1AspectTable: array[0..14] of Double = (
    1, 1, 0.6735, 0.7031, 0.7615, 0.8055, 0.8437, 0.8935,
    0.9815, 54/59 {PAL}, 1.0255, 1.0695, 11/10 {NTSC}, 1.1575, 1.2015
  );

procedure TimeProc(IDEvent, uReserved, dwUser, dwReserved1, dwReserved2: Integer); stdcall;
begin
  with TVideoPlayer(dwUser) do
  begin
    if Loaded and Assigned(OnTimer) then
      OnTimer;
  end;
end;

function TVideoPlayer.GetVolume: Integer;
begin
  if Loaded then
    Fiba.get_Volume(Result)
  else
    Result := 0;
end;

procedure TVideoPlayer.SetVolume(Value: Integer);
begin
  if Loaded then Fiba.put_Volume(Value);
end;

function TVideoPlayer.GetPosition: Integer;
var
  cp: Comp;
begin
  if Loaded then
  begin
    Fims.GetCurrentPosition(cp);
    if cp >= Duration then
    begin
      Fimc.Stop; // ̂ȁ[
      Mode := vpStop;
      if Assigned(OnEndPlay) then
        OnEndPlay;
    end;
    Result := Trunc(cp / 10000);
  end else
    Result := 0;
end;

procedure TVideoPlayer.SetPosition(Value: Integer);
var
  CurPos, StopPos: Comp;
begin
  if Loaded then
  begin
    CurPos := Value * 10000.0;
    if CurPos >= Duration then CurPos := Duration;
    StopPos := 0;
    Fims.SetPositions(CurPos, AM_SEEKING_AbsolutePositioning,
      StopPos, AM_SEEKING_NoPositioning);
    GetPosition;
  end;
end;

procedure TVideoPlayer.FreeInterface;
begin
  if Assigned(Fimc) then
    Fimc.Stop;
  if Assigned(Fivw) then
  begin
    Fivw.put_Visible(False);
    Fivw.put_Owner(OAHWND(0));
  end;
  Fimc := nil;
  Fivw := nil;
  Fims := nil;
  Figb := nil;
  Fibv := nil;
  //Fivf := nil;
  Fiba := nil;
  Fime := nil;
  Loaded := False;
end;

procedure TVideoPlayer.Play;
var
  cp: Comp;
begin
  if Loaded then
  begin
    Fims.GetCurrentPosition(cp);
    if (cp = Duration) and (Mode = vpStop) then
      Position := 0;
    Fimc.Run;
    Mode := vpPlay;
    if TimerID <> 0 then
    begin
      timeKillEvent(TimerID);
      TimerID := 0;
    end;
    TimerID := timeSetEvent(100, 50, TimeProc, DWORD(Self), TIME_PERIODIC);
  end;
end;

procedure TVideoPlayer.Stop;
begin
  if Loaded then
  begin
    Fimc.Stop;
    Position := 0;
    Fimc.StopWhenReady;
    Mode := vpStop;
  end;
end;

procedure TVideoPlayer.PauseOnly;
begin
  if Loaded then
  begin
    timeKillEvent(TimerID);
    TimerID := 0;
    Fimc.Pause;
    Mode := vpPause;
  end;
end;

procedure TVideoPlayer.Pause;
begin
  if Mode = vpPause then
    Play
  else
    PauseOnly;
end;

function TVideoPlayer.GetAVIHeader(var F: File): Boolean;
var
  size: Integer;
begin
  FillChar(AVIMainHeader, SizeOf(AVIMainHeader), 0);
  FillChar(BitmapInfoHeader, SizeOf(BitmapInfoHeader), 0);
  FillChar(WaveFormat, SizeOf(WaveFormat), 0);
  try
    if SearchChunk(F, 'RIFF', size) then
    begin
      SearchChunk(F, 'avih', size);
      BlockRead(F, AVIMainHeader, SizeOf(AVIMainHeader));
      SearchChunk(F, 'strf', size);
      Seek(F, FilePos(F)+8);
      BlockRead(F, BitmapInfoHeader, SizeOf(BitmapInfoHeader));
      SearchChunk(F, 'strf', size);
      Seek(F, FilePos(F)+8);
      BlockRead(F, WaveFormat, SizeOf(WaveFormat));
      Result := True;
    end else
      Result := False;
  except
    Result := False;
  end;
end;

procedure TVideoPlayer.GetMpegInfo(var F: File; var Aspect: Double; var w, h: Integer);
const
  bufsize = 8192;
var
  buf: string;
  a, len, i: Integer;
begin
  a := 0;
  Seek(F, 0);
  SetLength(buf, bufsize);
  while True do
  begin
    FillChar(buf[1], bufsize, 0);
    BlockRead(F, buf[1], bufsize, len);
    i := Pos(#$00#$00#$01#$b3, buf);
    if (i > 0) and (len > (i+8)) then
    begin
      w := Byte(buf[i+4]) shl 4 + Byte(buf[i+5]) shr 4;
      h := Byte(buf[i+5]) and $F + Byte(buf[i+6]);
      a := Byte(buf[i+7]) shr 4;
      Break;
    end;
    // 1MBǂ񂾂炠߂
    if Eof(F) or (FilePos(F) > 1024*1024) then
      Break;
    Seek(F, FilePos(F)-4);
  end;
  if (a > 14) or (a < 0) then a := 0;
  Aspect := Mpeg1AspectTable[a];
end;

procedure TVideoPlayer.GetFilterNames;
var
  ief: IEnumFilters;
  Filter: IBaseFilter;
  FilterInfo: TFilterInfo;
  f: ULONG;
  prop: ISpecifyPropertyPages;
begin
  FilterCount := 0;
  Figb.EnumFilters(ief);
  while ief.Next(1, Filter, f) = S_OK do
  begin
    Filter.QueryFilterInfo(FilterInfo);
    try
      prop := Filter as ISpecifyPropertyPages;
    except
      on EIntfCastError do Continue; // vpeBȂtB^͖
    end;
    FilterNames[FilterCount] := FilterInfo.achName;
    Inc(FilterCount);
  end;
end;

procedure TVideoPlayer.LoadFile(const FileName: string);
var
  tpf: Double;
  F: File;
  ws: WideString;
  mpegw, mpegh: Integer;
begin
  Loaded := False;
  Length := 0;
  Position := 0;
  Fps := 0;
  Mode := vpStop;

  if Assigned(Fimc) then
  begin
    Fimc.Stop;
    FreeInterface;
  end;

  FileMode := 0;
  AssignFile(F, FileName);
  try
    Reset(F, 1);
  except
    Exit;
  end;
  if GetAVIHeader(F) then
    Aspect := 1
  else
    GetMpegInfo(F, Aspect, mpegw, mpegh);
  CloseFile(F);

  try
    CoCreateInstance(CLSID_FilterGraph, nil, CLSCTX_INPROC_SERVER,
      IID_IGraphBuilder, Figb);
    ws := FileName;
    if Figb.RenderFile(PWideChar(ws), nil) <> S_OK then
    begin
      FreeInterface;
      Exit;
    end;

    Fivw := Figb as IVideoWindow;
    Fivw.put_Owner(OAHWND(DisplayWnd));
    Fivw.put_WindowStyle(WS_CHILD or WS_CLIPSIBLINGS);
    Fivw.put_MessageDrain(DisplayWnd);
    Fimc := Figb as IMediaControl;
    Fims := Figb as IMediaSeeking;
    Fime := Figb as IMediaEvent;
    //Fims.SetTimeFormat(TIME_FORMAT_FRAME);
    Fims.SetTimeFormat(TIME_FORMAT_MEDIA_TIME);
    Fibv := Figb as IBasicVideo;
    Fibv.GetVideoSize(Width, Height);
    // MPEGǂ̔
    if (Width <> mpegw) or (Height <> mpegh) then
      Aspect := 1;

    tpf := 0;
    Fibv.get_AvgTimePerFrame(tpf);
    if tpf = 0 then
    begin
      if AVIMainHeader.dwMicroSecPerFrame > 0 then
        Fps := 1000000 / AVIMainHeader.dwMicroSecPerFrame
      else
        Fps := 0;
    end else
      Fps := 1 / tpf;
    Fims.GetDuration(Duration);
    Length := Trunc(Duration / 10000);
    Self.FileName := FileName;
    Loaded := True;
    //Fivf := Figb as IVideoFrameStep;
    //Fims.SetRate(2.0);
    Fiba := Figb as IBasicAudio;

    GetFilterNames;
  except
    FreeInterface;
  end;
end;

procedure TVideoPlayer.SetWindowPosition(ALeft, ATop, AWidth, AHeight: Longint);
begin
  Fivw.SetWindowPosition(ALeft, ATop, AWidth, AHeight);
end;

function TVideoPlayer.WriteBitmapBits(const FileName: string): Boolean;
var
  ws: WideString;
  i, n: Integer;
  imd: IMediaDet;
  mt: TGUID;
begin
  Result := False;
  if Loaded then
  begin
    if FAILED(CoCreateInstance(CLSID_MediaDet, nil, CLSCTX_INPROC_SERVER, IID_IMediaDet, imd)) then
      Exit;
    ws := Self.FileName;
    if FAILED(imd.put_Filename(ws)) then
      Exit;
    if FAILED(imd.get_OutputStreams(n)) then
      Exit;
    for i := 0 to n - 1 do
    begin
      if FAILED(imd.put_CurrentStream(i)) then
        Continue;
      if FAILED(imd.get_StreamType(mt)) then
        Continue;
      if CompareMem(@mt, @MEDIATYPE_Video, SizeOf(mt)) then
      begin
        Result := imd.WriteBitmapBits(Position / 1000.0, Width, Height, FileName)= S_OK;
        Break;
      end;
    end;
  end;
end;

procedure TVideoPlayer.ShowFilterPropertyByName(const FilterName: string);
var
  Filter: IBaseFilter;
  prop: ISpecifyPropertyPages;
  pages: TCAGUID;
  ws: WideString;
  FilterInfo: TFilterInfo;
  ief: IEnumFilters;
  f: ULONG;
begin
  if Loaded then
  begin
    ws := FilterName;
    // FindFilterByName͂ȂWindows Media Source FilerŎs
    //{$C+}
    //if Figb.FindFilterByName(PWideChar(ws), Filter) <> S_OK then
    //  Assert(False);
    Figb.EnumFilters(ief);
    while ief.Next(1, Filter, f) = S_OK do
    begin
      Filter.QueryFilterInfo(FilterInfo);
      if WideString(FilterInfo.achName) = ws then
      begin;
        try
          prop := Filter as ISpecifyPropertyPages;
        except
          on EIntfCastError do Continue; // vpeBȂtB^͖
        end;
        if Assigned(prop) then
        begin
          prop.GetPages(pages);
          OleCreatePropertyFrame(
            DisplayWnd,          // eEBhE
            0,                   // x (\ς)
            0,                   // y (\ς)
            PWideChar(ws),       // _CAO {bNX̃LvV
            1,                   // tB^̐
            @Filter,             // tB^ւ̃|C^
            pages.cElems,        // vpeB y[W̐
            pages.pElems,        // vpeB y[W CLSID ւ̃|C^
            0,                   // P[ʎq
            0,                   // \ς
            nil                  // \ς
          );
          CoTaskMemFree(pages.pElems);
        end;
        Break;
      end;
    end;
  end;
end;

end.
